import json

from botocore.exceptions import ClientError
import tiros.util as util
from tiros.util import vprint

# The name for the IAM policy with permissions for Tiros to describe
# networks
TIROS_POLICY_NAME = 'Tiros'

# The minimal permissions Tiros requires to download a network snapshot
TIROS_POLICY = json.loads("""
{
    "Statement": [
        {
            "Action": [
                "ec2:Describe*",
                "elasticache:Describe*",
                "elasticloadbalancing:Describe*",
                "rds:Describe*"
            ],
            "Resource": "*",
            "Effect": "Allow"
        }
    ],
    "Version": "2012-10-17"
}
""")

# The policy that allows the Tiros account to assume a role in a
# customer account.
TIROS_ASSUME_ROLE_POLICY = json.loads("""
{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Action": "sts:AssumeRole",
            "Effect": "Allow",
            "Principal": {
                "AWS": "arn:aws:iam::748907555215:root"
            }
        }
    ]
}
""")

# The name for cross-account IAM roles to allow Tiros to make calls
# on the user's behalf.  Note that this name is actually required by
# the Tiros service, while iam_policy_name
TIROS_ROLE_NAME = 'Tiros'

# Policies to attach to the Lambda role
LAMBDA_POLICY_NAMES = [TIROS_POLICY_NAME,
                       'AWSLambdaBasicExecutionRole',
                       'AWSConfigRulesExecutionRole']

LAMBDA_ASSUME_ROLE_POLICY = json.loads("""
{
    "Version": "2012-10-17",
    "Statement": [
        {
            "Action": "sts:AssumeRole",
            "Effect": "Allow",
            "Principal": {
                "Service": "lambda.amazonaws.com"
            }
        }
    ]
}
""")

LAMBDA_ROLE_NAME = 'TirosLambda'


def s3_policy_statement(s3_prefix):
    s3_prefix = s3_prefix.replace('s3://', '')
    return {
        "Effect": "Allow",
        "Action": ["s3:PutObject", "s3:GetObject"],
        "Resource": ["arn:aws:s3:::" + s3_prefix + '*']
    }


def sns_policy_statement(sns_topic_arn):
    return {
        "Action": ["SNS:Publish"],
        "Effect": "Allow",
        "Resource": sns_topic_arn
    }


def lambda_inline_policy(s3_prefix, sns_topic_arn):
    statements = []
    if s3_prefix:
        statements += [s3_policy_statement(s3_prefix)]
    if sns_topic_arn:
        statements += [sns_policy_statement(sns_topic_arn)]
    return util.pretty({
        "Version": "2012-10-17",
        "Statement": statements
    })


def policy_arn(account, policy_name):
    return ''.join(['arn:aws:iam::', account, ':policy/', policy_name])


def role_arn(account, role_name):
    return ''.join(['arn:aws:iam::', account, ':role/', role_name])


def get_default_policy(client, account, policy_name):
    try:
        arn = policy_arn(account, policy_name)
        policy = client.get_policy(PolicyArn=arn)
        version = policy['Policy']['DefaultVersionId']
        doc = client.get_policy_version(PolicyArn=arn, VersionId=version)
        return doc['PolicyVersion']['Document']
    except ClientError:
        return None


def create_or_update_policy(client, account, policy_name, policy):
    existing_policy = get_default_policy(client, account, policy_name)
    if not existing_policy:
        vprint('Creating policy: ' + policy_name)
        client.create_policy(
            PolicyName=policy_name,
            PolicyDocument=util.pretty(policy))
    else:
        if existing_policy != policy:
            vprint('Updating policy: ' + policy_name)
            client.create_policy_version(
                PolicyArn=policy_arn(account, policy_name),
                PolicyDocument=util.pretty(policy),
                SetAsDefault=True)
        else:
            vprint('Policy is up to date: ' + policy_name)


def create_or_update_role(client, role_name, assume_role_policy):
    try:
        role = client.get_role(RoleName=role_name)
        policy = role['Role']['AssumeRolePolicyDocument']
        if policy != assume_role_policy:
            vprint('Updating role: ' + role_name)
            client.update_assume_role_policy(
                RoleName=role_name,
                PolicyDocument=util.pretty(assume_role_policy))
        else:
            vprint('Role is up to date: ' + role_name)
    except ClientError:
        vprint('Creating role: ' + role_name)
        client.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=util.pretty(assume_role_policy))


def add_policy_to_role(client, account, role_name, policy_name):
    policies = client.list_attached_role_policies(RoleName=role_name)
    if any([p['PolicyName'] == policy_name
            for p in policies['AttachedPolicies']]):
        vprint(role_name + ' policies are OK')
    else:
        try:
            arn = ''.join(['arn:aws:iam::', account, ':policy/', policy_name])
            client.attach_role_policy(RoleName=role_name, PolicyArn=arn)
            vprint('Added local policy: ' + policy_name + ' to role: ' + role_name)
        except ClientError:
            arn = 'arn:aws:iam::aws:policy/service-role/' + policy_name
            client.attach_role_policy(RoleName=role_name, PolicyArn=arn)
            vprint('Added managed policy: ' + policy_name + ' to role: ' + role_name)


def add_inline_policy_to_role(client, role_name, policy_name, policy):
    vprint('Adding inline policy: ' + policy_name + ' to role: ' + role_name)
    client.put_role_policy(
        RoleName=role_name,
        PolicyName=policy_name,
        PolicyDocument=policy)


def role_names(client):
    roles = client.list_roles()['Roles']
    return [r['RoleName'] for r in roles]


def delete_role(client, role_name):
    if role_name not in role_names(client):
        vprint('No such role: ' + role_name)
        return
    policies = client.list_attached_role_policies(RoleName=role_name)
    for p in policies.get('AttachedPolicies'):
        vprint('Detaching policy: ' + p.get('PolicyName') + ' from role: ' + role_name)
        client.detach_role_policy(RoleName=role_name,
                                  PolicyArn=p.get('PolicyArn'))
    policies = client.list_role_policies(RoleName=role_name)
    for p in policies.get('PolicyNames'):
        vprint('Detaching policy: ' + p + ' from role: ' + role_name)
        client.delete_role_policy(RoleName=role_name, PolicyName=p)
    vprint('Deleting role: ' + role_name)
    client.delete_role(RoleName=role_name)


def policy_names(client):
    policies = client.list_policies(Scope='Local')['Policies']
    return [p['PolicyName'] for p in policies]


def delete_policy(client, account, policy_name):
    if policy_name not in policy_names(client):
        vprint('No such policy: ' + policy_name)
        return
    vprint('Deleting policy: ' + policy_name)
    arn = policy_arn(account, policy_name)
    for v in client.list_policy_versions(PolicyArn=arn)['Versions']:
        if not v.get('IsDefaultVersion'):
            vid = v['VersionId']
            vprint('Deleting policy version: ' + vid)
            client.delete_policy_version(PolicyArn=arn, VersionId=vid)
    client.delete_policy(PolicyArn=arn)


def create_tiros_policy(session):
    client = session.client('iam')
    account = util.account(session)
    create_or_update_policy(client, account, TIROS_POLICY_NAME, TIROS_POLICY)


def delete_tiros_policy(session):
    client = session.client('iam')
    account = util.account(session)
    delete_policy(client, account, TIROS_POLICY_NAME)


def create_tiros_role(session):
    client = session.client('iam')
    account = util.account(session)
    create_or_update_role(client, TIROS_ROLE_NAME, TIROS_ASSUME_ROLE_POLICY)
    add_policy_to_role(client, account, TIROS_ROLE_NAME, TIROS_POLICY_NAME)


def delete_tiros_role(session):
    client = session.client('iam')
    delete_role(client, TIROS_ROLE_NAME)


def create_lambda_role(session, s3_prefix, sns_topic_name):
    client = session.client('iam')
    account = util.account(session)
    sns_topic_arn = 'arn:aws:sns:us-west-2:' + account + ':' + sns_topic_name
    create_or_update_role(client, LAMBDA_ROLE_NAME, LAMBDA_ASSUME_ROLE_POLICY)
    for p in LAMBDA_POLICY_NAMES:
        add_policy_to_role(client, account, LAMBDA_ROLE_NAME, p)
    if s3_prefix or sns_topic_name:
        add_inline_policy_to_role(
            client=client,
            role_name=LAMBDA_ROLE_NAME,
            policy_name='TirosLambdaS3SNS',
            policy=lambda_inline_policy(s3_prefix, sns_topic_arn))


def delete_lambda_role(session):
    client = session.client('iam')
    delete_role(client, LAMBDA_ROLE_NAME)



from __future__ import absolute_import, division, print_function, unicode_literals

import boto3
import json
import os
import sys

TIROS_VERSION = '1.0'

CONTENT_TYPE = 'application/json'
METHOD = 'POST'

VERBOSE = False


def vprint(x):
    if VERBOSE:
        print(str(x))


def eprint(s):
    print(s)
    sys.exit(1)


def file_contents(path):
    if os.path.exists(path):
        with open(path) as fd:
            return fd.read()
    else:
        # Don't use FileNotFoundError, which doesn't exist in Python2
        raise IOError('File not found: ' + path)


def quote(s):
    return '"' + s + '"'


def json_encoder(obj):
    """JSON serializer for objects not serializable by default json code"""
    if type(obj) == bytes:
        return obj.decode('utf-8')
    raise TypeError("Type not serializable")


def pretty(x):
    return json.dumps(x, indent=2, sort_keys=True, default=json_encoder)


def canonical(x):
    return json.dumps(x, sort_keys=True, default=json_encoder)


def assume_role_session(session, role_arn, region=None):
    """
    :param session: An existing boto Session
    :param role_arn: An IAM role to assume
    :param region: The region for the new Session, defaults to region of
    existing session.
    :return: A Session with the assumed role.
    """
    sts = session.client('sts')
    creds = sts.assume_role(
        RoleArn=role_arn,
        RoleSessionName='tiros',  # arbitrary
    )
    if 'Credentials' not in creds:
        eprint("Couldn't assume role: " + role_arn)
    cs = creds['Credentials']
    token = cs['SessionToken'] if 'SessionToken' in cs else None
    if not region:
        region = session.region_name
    return boto3.Session(
        aws_access_key_id=cs['AccessKeyId'],
        aws_secret_access_key=cs['SecretAccessKey'],
        aws_session_token=token,
        region_name=region)


def change_session_region(session, region):
    if region == session.region_name:
        return session
    creds = session.get_credentials()
    return boto3.Session(aws_access_key_id=creds.access_key,
                         aws_secret_access_key=creds.secret_key,
                         aws_session_token=creds.token,
                         region_name=region)


def account(session):
    return session.client('sts').get_caller_identity().get('Account')


# Fix input in python2
try:
    # noinspection PyShadowingBuiltins
    safe_input = raw_input
except NameError:
    safe_input = input



import json
import tiros.iam as iam
import tiros.util as util
from botocore.exceptions import ClientError
from tiros.util import vprint

CODE_BUCKET = 'tiros-lambda'
CODE_KEY = 'tiros-lambda.zip'
CONFIG_ACTION = 'lambda:InvokeFunction'
CONFIG_S3_PREFIX_KEY = 's3-prefix'
CONFIG_PRINCIPAL = 'config.amazonaws.com'
CONFIG_RULE_NAME = 'TirosLambda'
FUN_DESCRIPTION = 'Tiros Lambda'
FUN_HANDLER = 'tiros_lambda.lambda_handler'
FUN_NAME = 'TirosLambda'
FUN_RUNTIME = 'python2.7'
FUN_TIMEOUT = 300
QUERIES_FILE_NAME = 'queries.tql'


def config_s3_prefix_key():
    return CONFIG_S3_PREFIX_KEY


def functions(client):
    res = client.list_functions()
    funs = [d['FunctionName'] for d in res['Functions']]
    m = res.get('NextMarker')
    while m:
        res = client.list_functions(Marker=m)
        funs += [d['FunctionName'] for d in res['Functions']]
        m = res.get('NextMarker')
    return funs


def rule_names(config_client):
    rules = config_client.describe_config_rules()
    return [r['ConfigRuleName'] for r in rules['ConfigRules']]


def lambda_function_exists(client):
    return FUN_NAME in functions(client)


def lambda_arn(account):
    return ''.join(
        ['arn:aws:lambda:us-west-2:', account, ':function:', FUN_NAME])


def create_or_update_lambda_function(session):
    client = session.client('lambda')
    account = util.account(session)
    role = iam.role_arn(account, iam.LAMBDA_ROLE_NAME)
    if lambda_function_exists(client):
        vprint('Updating existing lambda: ' + FUN_NAME)
        client.update_function_code(
            FunctionName=FUN_NAME,
            S3Bucket=CODE_BUCKET,
            S3Key=CODE_KEY)
    else:
        vprint('Creating new lambda: ' + FUN_NAME)
        client.create_function(
            FunctionName=FUN_NAME,
            Runtime=FUN_RUNTIME,
            Role=role,
            Handler=FUN_HANDLER,
            Code={
                'S3Bucket': CODE_BUCKET,
                'S3Key': CODE_KEY
            },
            Description=FUN_DESCRIPTION,
            Timeout=FUN_TIMEOUT)


def delete_lambda(session):
    client = session.client('lambda')
    if lambda_function_exists(client):
        vprint('Deleting lambda: ' + FUN_NAME)
        client.delete_function(FunctionName=FUN_NAME)
    else:
        vprint('There is no lambda named: ' + FUN_NAME)


def create_rule(session, region, s3_prefix, sns_topic_name):
    config_client = session.client('config')
    lambda_client = session.client('lambda')
    if not region:
        region = session.region_name
    params = {
        'region': region,
        's3-prefix': s3_prefix
    }
    if sns_topic_name:
        params['sns-topic-name'] = sns_topic_name
    account = util.account(session)
    rules = rule_names(config_client)
    if CONFIG_RULE_NAME in rules:
        vprint('Rule already created')
        return
    vprint('Giving Config permission to call the Lambda')
    try:
        lambda_client.add_permission(
            FunctionName=FUN_NAME,
            StatementId='ConfigPermission',
            Action=CONFIG_ACTION,
            Principal=CONFIG_PRINCIPAL,
            SourceAccount=account
        )
    except ClientError:
        # If we've already added permission this raises an exception
        # because the StatementId is a duplicate
        pass
    vprint('Creating Config Rule: ' + CONFIG_RULE_NAME)
    try:
        config_client.put_config_rule(
            ConfigRule={
                'ConfigRuleName': CONFIG_RULE_NAME,
                'Source': {
                    'Owner': 'CUSTOM_LAMBDA',
                    'SourceIdentifier': lambda_arn(account),
                    'SourceDetails': [
                        {
                            'EventSource': 'aws.config',
                            'MessageType': 'ScheduledNotification'
                        },
                    ]
                },
                # 'MaximumExecutionFrequency': 'One_Hour',
                'InputParameters': json.dumps(params)
            }
        )
    except ClientError as exn:
        # An error if Config is not set up refers to the deliveryFrequency
        s = str(exn)
        if 'deliveryFrequency' in s and 'MaximumExecutionFrequency' in s:
            print('Got an error suggesting Config has not been turned on.')
            print('Please enable Config before creating a Config Rule from the Tiros CLI.')
            exit(1)


def delete_rule(session):
    client = session.client('config')
    rules = rule_names(client)
    if CONFIG_RULE_NAME not in rules:
        vprint('No Config Rule named: ' + CONFIG_RULE_NAME)
        return
    vprint('Deleting Config Rule: ' + CONFIG_RULE_NAME)
    client.delete_config_rule(ConfigRuleName=CONFIG_RULE_NAME)
